
/* 
   Project:     SOFTSUSY 
   File:        linalg.h
   Author:      Ben Allanach 
   Manual:      hep-ph/0104145, Comp. Phys. Comm. 143 (2002) 305 
   Webpage:     http://allanach.home.cern.ch/allanach/softsusy.html
   Description: DoubleVector and DoubleMatrix classes of doubles and 
                operations between them, complexified copies also

   $Log: linalg.h,v $
   Revision 1.9  2002/09/23 18:13:50  allanach
   Eigenvalue order with non symmetric diagonalisation setx

   Revision 1.8  2002/07/19 15:54:06  allanach
   SOFTSUSY1.5 version

   Revision 1.7  2002/06/10 16:13:07  allanach
   Speeded things up by adding option for array bounds checking

   Revision 1.5  2002/04/30 16:23:59  allanach
   Added matrix inversion routine

   Revision 1.4  2001/10/03 13:34:16  allanach
   Changed name of complex.h to avoid conflict with STD libraries

   Revision 1.3  2001/07/18 14:42:51  allanach
   Added proper header info
*/

#ifndef LINALG_H
#define LINALG_H

#include <stdlib.h>
#include <iostream.h>
#include <math.h>
#include <def.h>
#include "utils.h" 
#include "mycomplex.h"

inline void freeMyDoubleVector(double *v, long nl, long nh) {
  free((char *)(v+nl-1)); 
};

// DoubleVector is incomplete: add more operator overloads if required
class DoubleVector {
private:
  int start, end;
  double * x;
  
  // Memory handling routines
  double *myDoubleVector(const long nl, const long nh);
  void setVec(const double *t, long st, long nd);

public:
  DoubleVector(int e);
  DoubleVector(int s, int e);
  inline DoubleVector(const DoubleVector &v);
  ~DoubleVector();

  void setEnd(int e);
  const DoubleVector & operator=(const DoubleVector &v);
  DoubleVector operator+(const DoubleVector &v);
  DoubleVector operator-(const DoubleVector &v);
  DoubleVector operator*(const DoubleVector &v);
  DoubleVector operator*(double f);
  double & operator() (int i);
  double dot(const DoubleVector & v) const;
  inline double display(int i) const { return x[i]; };
  int displayStart() const { return start; };
  int displayEnd() const { return end; };
  DoubleVector display() const { return *this; };
  void display(double * a) const;

  DoubleVector apply(double (*fn)(double));
  void set(int i, double f);
  double max() const;
  double min(int & p) const;
  
  void swap(int i, int j);
};


DoubleVector operator*(double f, const DoubleVector & V);
ostream & operator <<(ostream &left, const DoubleVector &V);
istream & operator>>(istream & left, DoubleVector &V);

inline DoubleVector::~DoubleVector() { freeMyDoubleVector(x, start, end); }

inline DoubleVector::DoubleVector(const DoubleVector &v)
  : start(v.start), end(v.end)
{ setVec(v.x, v.start, v.end); } 

inline double & DoubleVector::operator() (int i) {
#ifdef ARRAY_BOUNDS_CHECKING
  if (i>end || i<start) {
      cout << "Trying to access " << i << "th element of DoubleVector\n";
      cout << "start " << start << " end " << end;
      cout << *this;
    }
#endif
  return x[i];
}

inline void DoubleVector::set(int i, double f) {
#ifdef ARRAY_BOUNDS_CHECKING
  if (i < start || i > end) 
    {
      cout << "Cannot access " << i << "th element of vector " << *this;
      exit(1);
    }
#endif
  x[i] = f;
}

/* -----------------------------------------------------*/

/* free a double matrix allocated by myDoubleMatrix() */
inline void freeMyDoubleMatrix(double **m, long nrl, long nrh, long ncl, long
		   nch) {
  free((char *) (m[nrl]+ncl-1));
  free((char *) (m+nrl-1));
}


class DoubleMatrix; class ComplexMatrix;
inline DoubleMatrix operator-(const DoubleMatrix & A);
//DoubleMatrix operator*(const DoubleMatrix & A, const DoubleMatrix & V);
DoubleMatrix operator*(double f, const DoubleMatrix &M);
//DoubleVector operator*(const DoubleMatrix & M, const DoubleVector & V);
DoubleVector operator*(const DoubleVector & V, const DoubleMatrix & M);
DoubleMatrix operator-(double f, const DoubleMatrix &M);
DoubleMatrix operator+(double f, const DoubleMatrix &M);
DoubleMatrix outerProduct(const DoubleVector & v1, const DoubleVector & v2);
istream & operator>>(istream & left, DoubleMatrix &M);
ostream & operator <<(ostream &left, const DoubleMatrix &V);
DoubleMatrix rot2d(double theta);

// DoubleMatrix from 1..rows, 1..cols
class DoubleMatrix {
private:
  int rows, cols;
  double **x;
  
  double **myDoubleMatrix(long nrl, long nrh, long ncl, long nch) const;
  double *myDoubleVector(const long nl, const long nh) const;
  void setMat(double **t, int r, int c);
  
public:
  DoubleMatrix(int r, int c);
  inline DoubleMatrix(const DoubleMatrix &m);
  
  // Makes diagonal square matrix out of W
  DoubleMatrix(const DoubleVector &v);  
  ~DoubleMatrix();
  
  int displayRows() const { return rows; };
  int displayCols() const { return cols; };

  const DoubleMatrix & operator=(const DoubleMatrix &v);  
  const DoubleMatrix & operator=(const double &v);  
  DoubleMatrix operator+(const DoubleMatrix &v) const;  
  DoubleMatrix operator-(const DoubleMatrix &v) const;  
  DoubleMatrix operator+(double f) const;  
  DoubleMatrix operator-(double f) const;  
  DoubleMatrix operator*(const DoubleMatrix &v) const;  
  DoubleMatrix operator*(const double f) const;
  double trace() const;  
  DoubleVector operator*(const DoubleVector &) const;
  DoubleMatrix operator/(const double f) const;
  double & operator() (int i, int j);
  DoubleMatrix transpose() const;

  void swaprows(int i, int j);
  void swapcols(int i,int j);
  void associateOrderAbs(DoubleVector &v);
  double display(int i, int j) const;
  DoubleMatrix display() const { return *this; };

  double sumElements() const;  
  void symmetrise();
  double compare(const DoubleMatrix & a) const;
  double compare(const ComplexMatrix & a) const;
  DoubleVector diagVals() const;
  double diagonalise(DoubleMatrix & u, DoubleMatrix & v, DoubleVector & w)
    const;
  double diagonaliseSym (ComplexMatrix & v, DoubleVector & w) const;
  double diagonaliseSym(DoubleMatrix & v, DoubleVector & w) const;
  DoubleMatrix inverse() const;
  DoubleVector sym2by2(double & theta) const;
  DoubleMatrix apply(double (*fn)(double));
  DoubleVector asy2by2(double & thetaL, double & thetaR) const;
  double min(int & k, int & l) const;
  void displayMat(double ** temp) const;  
  void associateOrderAbs(DoubleMatrix & u, DoubleMatrix & v, DoubleVector &w) const;
};

inline DoubleMatrix::DoubleMatrix(const DoubleMatrix &m) 
  : rows(m.rows), cols(m.cols) { setMat(m.x, rows, cols); }

inline DoubleMatrix::~DoubleMatrix() { freeMyDoubleMatrix(x, 1, rows, 1, cols); }

inline DoubleMatrix DoubleMatrix::operator/(const double f) const {
  return *this * (1.0 / f);
}

inline double & DoubleMatrix::operator() (int i, int j) { 
#ifdef ARRAY_BOUNDS_CHECKING
  if (i > rows || j > cols || i < 1 || j < 1) { 
    cout << "Trying to access " << i << "," << j << 
      "th element of DoubleMatrix\n" << *this;
    exit(1); 
  }
#endif
  return x[i][j]; 
}  

inline double DoubleMatrix::display(int i, int j) const { 
#ifdef ARRAY_BOUNDS_CHECKING
  if (i < 1 || i > displayRows() || j < 1 || j > displayCols()) { 
    cout << "Error: Requested display (" << i << "," << j << 
	")th element of matrix " << *this; 
    exit(1); 
    }
#endif
  return x[i][j]; 
};  

inline DoubleMatrix operator-(const DoubleMatrix & A) { return -1.0 * A; }

// Diagonalisation routines
void positivise(double thetaL, double thetaR, const DoubleVector & diag,
		  ComplexMatrix & u, ComplexMatrix & v);
void diagonaliseSvd(DoubleMatrix & a, DoubleVector & w, DoubleMatrix & v);
double pythagoras(double a, double b);
void diagonaliseJac(DoubleMatrix & a,  int n,  DoubleVector & d, DoubleMatrix
		    & v,  int *nrot);

/* ------------------------Complex classes-------------------- */

inline void freeMyComplexVector(Complex *v, long nl, long nh) {
  free((char *)(v+nl-1)); 
};

// ComplexVector is incomplete: add more operator overloads if required
class ComplexVector {
private:
  int start, end;
  Complex * x;
  
  // Memory handling routines
  Complex *myComplexVector(const long nl, const long nh);
  void setVec(const Complex *t, long st, long nd);

public:
  ComplexVector(int e);
  ComplexVector(int s, int e);
  inline ComplexVector(const ComplexVector &v);
  ~ComplexVector();

  void setEnd(int e);
  ComplexVector & operator=(const ComplexVector &v);
  ComplexVector operator+(const ComplexVector &v);
  ComplexVector operator-(const ComplexVector &v);
  ComplexVector operator*(const ComplexVector &v);
  ComplexVector operator*(Complex f);
  Complex & operator() (int i);
  Complex dot(const ComplexVector & v) const;
  inline Complex display(int i) const { return x[i]; };
  int displayStart() const { return start; };
  int displayEnd() const { return end; };
  ComplexVector display() const { return *this; };

  ComplexVector apply(Complex (*fn)(Complex));
  void set(int i, Complex f);
  Complex max() const;
  Complex min(int & p) const;
  
  void swap(int i, int j);
};

ComplexVector operator*(double f, const ComplexVector & V);
ComplexVector operator*(Complex f, const ComplexVector & V);
ostream & operator <<(ostream &left, const ComplexVector &V);
istream & operator>>(istream & left, ComplexVector &V);

inline ComplexVector::~ComplexVector() { freeMyComplexVector(x, start, end); }

inline ComplexVector::ComplexVector(const ComplexVector &v)
  : start(v.start), end(v.end)
{ setVec(v.x, v.start, v.end); } 

inline Complex & ComplexVector::operator() (int i) {
#ifdef ARRAY_BOUNDS_CHECKING
  if (i>end || i<start) {
      cout << "Trying to access " << i << "th element of DoubleVector\n";
      cout << "start " << start << " end " << end;
      cout << *this;
    }
#endif
  return x[i];
}

inline void ComplexVector::set(int i, Complex f) {
#ifdef ARRAY_BOUNDS_CHECKING
  if (i < start || i > end) 
    {
      cout << "Cannot access " << i << "th element of vector " << *this;
      exit(1);
    }
#endif
  x[i] = f;
}

/* -----------------------------------------------------*/

/* free a Complex matrix allocated by myComplexMatrix() */
inline void freeMyComplexMatrix(Complex **m, long nrl, long nrh, long ncl, long
		   nch) {
  free((char *) (m[nrl]+ncl-1));
  free((char *) (m+nrl-1));
}


class ComplexMatrix;
ComplexMatrix operator*(const ComplexMatrix & A, const ComplexMatrix & V);
ComplexMatrix operator*(Complex f, const ComplexMatrix &M);
ComplexMatrix operator*(const DoubleMatrix & A, const ComplexMatrix &M);
ComplexMatrix operator*(double f, const ComplexMatrix &M);
ComplexVector operator*(const ComplexMatrix & M, const ComplexVector & V);
ComplexVector operator*(ComplexMatrix & M, DoubleVector & V);
ComplexVector operator*(const ComplexVector & V, const ComplexMatrix & M);
ComplexMatrix operator-(Complex f, const ComplexMatrix &M);
ComplexMatrix operator+(Complex f, const ComplexMatrix &M);
ComplexMatrix outerProduct(const ComplexVector & v1, const ComplexVector &
			   v2);
ComplexVector operator*(DoubleMatrix & m, ComplexVector & v);
istream & operator>>(istream & left, ComplexMatrix &M);
ostream & operator <<(ostream &left, const ComplexMatrix &V);

// ComplexMatrix from 1..rows, 1..cols
class ComplexMatrix {
private:
  int rows, cols;
  Complex **x;
  
  Complex **myComplexMatrix(long nrl, long nrh, long ncl, long nch) const;
  Complex *myComplexVector(const long nl, const long nh) const;
  void setMat(Complex **t, int r, int c);
  
public:
  ComplexMatrix(int r, int c);
  inline ComplexMatrix(const ComplexMatrix &m);
  ComplexMatrix(const DoubleVector &v);
  
  // Makes diagonal square matrix out of W
  ComplexMatrix(const ComplexVector &v);  
  ~ComplexMatrix();
  
  int displayRows() const { return rows; };
  int displayCols() const { return cols; };

  const ComplexMatrix & operator=(const ComplexMatrix &v);  
  const ComplexMatrix & operator=(const Complex &v);  
  ComplexMatrix operator+(const ComplexMatrix &v);  
  ComplexMatrix operator-(const ComplexMatrix &v);  
  ComplexMatrix operator+(Complex f);  
  ComplexMatrix operator-(Complex f);  
  ComplexMatrix operator*(const ComplexMatrix &v);  
  ComplexMatrix operator*(const Complex f);
  Complex trace() const;  
  ComplexVector operator*(const ComplexVector &);
  ComplexMatrix operator/(const Complex f);
  ComplexVector operator*(const DoubleVector & v);

  Complex & operator() (int i, int j);
  ComplexMatrix transpose() const;
  ComplexMatrix hermitianConjugate() const;
  ComplexMatrix complexConjugate() const;

  ComplexMatrix operator*(const DoubleMatrix &v);

  void swaprows(int i, int j);
  void swapcols(int i,int j);
  void associateOrderAbs(ComplexVector &v);
  Complex display(int i, int j) const;
  ComplexMatrix display() const { return *this; };

  Complex sumElements() const;  
  void symmetrise();
  double compare(const ComplexMatrix & a) const;
  Complex min(int & k, int & l) const;
  void displayMat(Complex ** temp) const;  
};

inline ComplexMatrix::ComplexMatrix(const ComplexMatrix &m) 
  : rows(m.rows), cols(m.cols) { setMat(m.x, rows, cols); }

inline ComplexMatrix::~ComplexMatrix() { freeMyComplexMatrix(x, 1, rows, 1, cols); }

inline ComplexMatrix ComplexMatrix::operator/(const Complex f) {
  return *this * (1.0 / f);
}

inline Complex & ComplexMatrix::operator() (int i, int j) { 
#ifdef ARRAY_BOUNDS_CHECKING
  if (i > rows || j > cols || i < 0 || j < 0) { 
    cout << "Trying to access " << i << "," << j << 
      "th element of ComplexMatrix\n" << *this;
    exit(1); 
  }
#endif
  return x[i][j]; 
}  

inline Complex ComplexMatrix::display(int i, int j) const { 
  if (i < 1 || i > displayRows() || j < 1 || j > displayCols()) { 
    cout << "Error: Requested display (" << i << "," << j << 
	")th element of matrix " << *this; 
    exit(1); 
    }
  return x[i][j]; 
};  

#endif
