
/* 
   Project:     SOFTSUSY 
   File:        lowe.h
   Author:      Ben Allanach 
   Manual:      hep-ph/0104145, Comp. Phys. Comm. 143 (2002) 305 
   Webpage:     http://allanach.home.cern.ch/allanach/softsusy.html
   Description: QedQcd object contains Standard Model quark and lepton masses.
                It integrates them using 3 loop qcd x 1 loop qed 
		effective theory.

   $Log: lowe.h,v $
   Revision 1.5  2002/09/23 18:13:50  allanach
   Eigenvalue order with non symmetric diagonalisation setx

   Revision 1.4  2002/09/20 15:37:10  allanach
   Adding quark-mixing routines

   Revision 1.3  2002/07/30 12:57:31  allanach
   SOFTSUSY1.5

   Revision 1.5  2002/02/20 17:20:02  allanach
   Treatment of bottom quark mass input changed. Now can input mb(mb) and/or
   pole(mb). If only one is input, with the other being a question mark, it
   will be calculated to three loop QCD from the input.

   Revision 1.4  2002/02/18 19:18:25  allanach
   Input is now bottom POLE mass. Running MSbar mass is now calculated!

   Revision 1.3  2001/07/18 14:42:51  allanach
   Added proper header info
*/

#ifndef LOWE_H
#define LOWE_H

#include<iostream.h>
#include<math.h>
#include<fstream.h>
#include<string.h>
#include<linalg.h>
#include<rge.h>

#include<def.h>

typedef enum {mUp=1, mCharm, mTop, mDown, mStrange, mBottom, mElectron,
	      mMuon, mTau} mass;
typedef enum {ALPHA=1, ALPHAS} leGauge;

DoubleVector gaugeDerivs(double, const DoubleVector &);

class QedQcd;
/* --------------RGELowe starts here------------ */
class QedQcd: public RGE
{
private:
  DoubleVector a;   // gauge couplings
  DoubleVector mf;  // fermion running masses
  double mtPole, mbPole;
  
public:
  QedQcd();
  QedQcd(const QedQcd &);
  const QedQcd& operator=(const QedQcd & m);
  virtual ~QedQcd() {};
  
  void setPoleMt(double mt) { mtPole = mt; };
  void setPoleMb(double mb) { mbPole = mb; };
  void setMass(mass mno, double m) { mf(mno) = m; };
  void setAlpha(leGauge ai, double ap) { a(ai) = ap; }; 
  void set(const DoubleVector &);     // For exporting beta functions to Runge-Kutta
  
  double displayPoleMt() const { return mtPole; };
  double displayPoleMb() const { return mbPole; };
  DoubleVector displayMass() const { return mf; };
  double displayMass(mass mno) const { return mf.display(mno); };
  double displayAlpha(leGauge ai) const { return a.display(ai); };
  DoubleVector display() const;
  
  int flavours(double) const;  // returns number of active flavours
  
  // Individual contributions to beta functions: alphaS, alphaEm, masses
  double qedBeta() const;
  double qcdBeta() const;
  void massBeta(DoubleVector &) const;
  DoubleVector beta() const;
  
  // Does not run the masses
  void runGauge(double, double);
  double extractPoleMb(double asMb);
  double extractRunningMb(double asMb);
  void calcRunningMb();
  void calcPoleMb();

  // Evolves object to running top mass
  void toMt();
  void toMz();
  DoubleVector  getGaugeMu(const double m2, const
		     double sinth) const;
};

// Input numbers into the object: by file stream
ostream & operator <<(ostream &, const QedQcd &);
istream & operator >>(istream &left, QedQcd &m);
void readIn(QedQcd &, char [80]); // or from a file

double getRunMt(double poleMt, double asmt);
double getAsmt(double mtop, double alphasMz);
double getRunMtFromMz(double poleMt, double asMZ);

inline QedQcd::QedQcd(const QedQcd &m)
  : a(m.a), mf(m.mf), mtPole(m.mtPole), mbPole(m.mbPole) { 
  setPars(11); 
  setMu(m.displayMu());
  setLoops(m.displayLoops());
  setThresholds(m.displayThresholds());
}

void massFermions(const QedQcd & r, DoubleMatrix & mDon, 
		  DoubleMatrix & mUpq, DoubleMatrix & mEle);
void doQuarkMixing(DoubleMatrix & vCkm, DoubleMatrix & mDon, 
		 DoubleMatrix & mUpq);
#endif

